"""Module for searchcontext for collection of cases."""

from ._search_context import SearchContext


class Cases(SearchContext):
    def __init__(self, sc, uuids):
        super().__init__(sc._sumo, must=[{"ids": {"values": uuids}}])
        self._hits = uuids
        return

    def _maybe_prefetch(self, index):
        return

    async def _maybe_prefetch_async(self, index):
        return

    def filter(self, **kwargs):
        sc = super().filter(**kwargs)
        uuids = sc.get_field_values("fmu.case.uuid.keyword")
        return Cases(sc, uuids)



"""context  pytest"""

import sys
from pathlib import Path

from fmu.sumo.explorer import Explorer
from fmu.sumo.explorer.objects._document import Document
from fmu.sumo.explorer.objects._search_context import SearchContext
from fmu.sumo.explorer.objects.case import Case
from fmu.sumo.explorer.objects.polygons import Polygons
from fmu.sumo.explorer.objects.surface import Surface
from fmu.sumo.explorer.objects.table import Table


def add_path():
    """Way to add package path to sys.path for testing"""
    # Adapted from https://docs.python-guide.org/writing/structure/
    # Turned into function because the details here didn't work
    package_path = str(Path(__file__).parent.absolute() / "../src/")
    while package_path in sys.path:
        sys.path.remove(package_path)
    sys.path.insert(0, package_path)


add_path()



"""Module containing class for cube object"""

from typing import Dict, Tuple

from sumo.wrapper import SumoClient

from ._child import Child


class Cube(Child):
    """Class representig a seismic cube object in Sumo"""

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata (dict): cube metadata
        """
        super().__init__(sumo, metadata, blob)

    def _extract_auth(self, res) -> Tuple[str, str]:
        try:
            res = res.json()
            url = res.get("baseuri") + self.uuid
            sas = res.get("auth")
        except Exception:
            url, sas = res.text.split("?")
            pass
        return url, sas

    @property
    def auth(self) -> Tuple[str, str]:
        res = self._sumo.get(f"/objects('{self.uuid}')/blob/authuri")
        return self._extract_auth(res)

    @property
    async def auth_async(self) -> Tuple[str, str]:
        res = await self._sumo.get_async(
            f"/objects('{self.uuid}')/blob/authuri"
        )
        return self._extract_auth(res)

    @property
    def openvds_handle(self):
        try:
            import openvds
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import openvds; probably not installed."
            )

        url, sas = self.auth
        url = url.replace("https://", "azureSAS://") + "/"
        sas = "Suffix=?" + sas
        return openvds.open(url, sas)

    @property
    async def openvds_handle_async(self):
        try:
            import openvds
        except ModuleNotFoundError:
            raise RuntimeError(
                "Unable to import openvds; probably not installed."
            )

        url, sas = await self.auth_async
        url = url.replace("https://", "azureSAS://") + "/"
        sas = "Suffix=?" + sas
        return openvds.open(url, sas)



"""Module containing class for exploring results from sumo"""

import warnings
from typing import Optional

import httpx
from sumo.wrapper import SumoClient

from .objects._search_context import SearchContext
from .objects.cases import Cases


class Explorer(SearchContext):
    """Class for consuming FMU results from Sumo.
    The Sumo Explorer is a Python package for consuming FMU results stored
    in Sumo. It is FMU aware, and creates an abstraction on top of the
    Sumo API. The purpose of the package is to create an FMU-oriented
    Python interface towards FMU data in Sumo, and make it easy for FMU
    users in various contexts to use data stored in Sumo.

    Examples of use cases:
      - Applications (example: Webviz)
      - Scripts (example: Local post-processing functions)
      - Manual data browsing and visualization (example: A Jupyter Notebook)
    """

    def __init__(
        self,
        env: str = "prod",
        token: Optional[str] = None,
        interactive: bool = True,
        keep_alive: Optional[str] = None,
        http_client=None,
        async_http_client=None,
    ):
        """Initialize the Explorer class

        Args:
            env (str): Sumo environment
            token (str): authenticate with existing token
            interactive (bool): authenticate using interactive flow (browser)
            keep_alive (str): point in time lifespan (deprecated and ignored)
        """
        sumo = SumoClient(
            env,
            token=token,
            interactive=interactive,
            timeout=httpx.Timeout(180.0),
            http_client=http_client,
            async_http_client=async_http_client,
        )
        SearchContext.__init__(self, sumo)
        if keep_alive:
            warnings.warn(
                "The constructor argument 'keep_alive' to class 'Explorer' has been deprecated.",
                DeprecationWarning,
            )

    @property
    def cases(self):
        uuids = self._context_for_class("case").uuids
        return Cases(self, uuids)

    def get_permissions(self, asset: Optional[str] = None):
        """Get permissions

        Args:
            asset (str): asset in Sumo

        Returns:
          dict: Dictionary of user permissions
        """
        res = self._sumo.get("/userpermissions").json()

        if asset is not None and asset not in res:
            raise PermissionError(f"No permissions for asset: {asset}")

        return res

    async def get_permissions_async(self, asset: Optional[str] = None):
        """Get permissions

        Args:
            asset (str): asset in Sumo

        Returns:
          dict: Dictionary of user permissions
        """
        res = await self._sumo.get_async("/userpermissions")
        res = res.json()

        if asset is not None and asset not in res:
            raise PermissionError(f"No permissions for asset: {asset}")

        return res



"""
Complex(ish) filters for use with fmu-sumo Explorer.
"""


class Filters:
    # Filter that matches 4d-seismic objects.
    seismic4d = {
        "bool": {
            "must": [
                {"term": {"data.content.keyword": "seismic"}},
                {"exists": {"field": "data.time.t0.label"}},
                {"exists": {"field": "data.time.t1.label"}},
            ]
        }
    }

    # Filter that matches aggregations
    aggregations = {"exists": {"field": "fmu.aggregation.operation"}}

    # Filter that matches observations
    observations = {
        "bool": {
            "must_not": [
                {"exists": {"field": "fmu.iteration.name.keyword"}},
                {"exists": {"field": "fmu.realization.id"}},
            ]
        }
    }

    # Filter that matches realizations
    realizations = {"exists": {"field": "fmu.realization.id"}}



"""Module for searchcontext for collection of iterations."""

from typing import Dict, List

from ._search_context import SearchContext


class Iterations(SearchContext):
    def __init__(self, sc, uuids):
        super().__init__(sc._sumo, must=[{"ids": {"values": uuids}}])
        self._hits = uuids
        return

    @property
    def classes(self) -> List[str]:
        return ["iteration"]

    @property
    async def classes_async(self) -> List[str]:
        return ["iteration"]

    def _maybe_prefetch(self, index):
        return

    async def _maybe_prefetch_async(self, index):
        return

    def get_object(self, uuid: str) -> Dict:
        """Get metadata object by uuid

        Args:
            uuid (str): uuid of metadata object
            select (List[str]): list of metadata fields to return

        Returns:
            Dict: a metadata object
        """
        obj = self._cache.get(uuid)
        if obj is None:
            obj = self.get_iteration_by_uuid(uuid)
            self._cache.put(uuid, obj)
            pass

        return obj

    async def get_object_async(self, uuid: str) -> Dict:
        """Get metadata object by uuid

        Args:
            uuid (str): uuid of metadata object
            select (List[str]): list of metadata fields to return

        Returns:
            Dict: a metadata object
        """

        obj = self._cache.get(uuid)
        if obj is None:
            obj = await self.get_iteration_by_uuid_async(uuid)
            self._cache.put(uuid, obj)

        return obj

    def filter(self, **kwargs):
        sc = super().filter(**kwargs)
        uuids = sc.get_field_values("fmu.iteration.uuid.keyword")
        return Iterations(sc, uuids)



"""Top-level package for fmu.sumo.explorer"""

try:
    from ._version import version

    __version__ = version
except ImportError:
    __version__ = "0.0.0"


from fmu.sumo.explorer.explorer import Explorer
from fmu.sumo.explorer.filters import Filters
from fmu.sumo.explorer.timefilter import TimeFilter, TimeType



"""Test table objects.

* Table
* AggregatedTable
* TableCollection

"""

import pandas as pd
import pyarrow as pa
import pytest

from fmu.sumo.explorer import Explorer

# Fixed test case ("Drogon_AHM_2023-02-22") in Sumo/DEV
TESTCASE_UUID = "10f41041-2c17-4374-a735-bb0de62e29dc"


@pytest.fixture(name="explorer")
def fixture_explorer(token: str) -> Explorer:
    """Returns explorer"""
    return Explorer("dev", token=token)


@pytest.fixture(name="case")
def fixture_case(explorer: Explorer):
    """Return fixed testcase."""
    return explorer.get_case_by_uuid(TESTCASE_UUID)


@pytest.fixture(name="table")
def fixture_table(case):
    """Get one table for further testing."""
    return case.tables[0]


### Table


def test_table_to_pandas(table):
    """Test the to_pandas method."""
    df = table.to_pandas()
    assert isinstance(df, pd.DataFrame)


def test_table_to_arrow(table):
    """Test the to_arrow() method"""
    arrow = table.to_arrow()
    assert isinstance(arrow, pa.Table)



"""Module containing class for dictionary object"""

import json
from typing import Dict, Optional

from sumo.wrapper import SumoClient

from fmu.sumo.explorer.objects._child import Child


class Dictionary(Child):
    """Class representing a dictionary object in Sumo"""

    _parsed: Optional[Dict]

    def __init__(self, sumo: SumoClient, metadata: Dict, blob=None) -> None:
        """
        Args:
            sumo (SumoClient): connection to Sumo
            metadata (dict): dictionary metadata
        """
        self._parsed: Optional[Dict] = None

        super().__init__(sumo, metadata, blob)

    def parse(self) -> Dict:
        parsed = (
            json.loads(self.blob.read().decode("utf-8"))
            if self._parsed is None
            else self._parsed
        )
        if self._parsed is None:
            self._parsed = parsed
        return parsed

    async def parse_async(self) -> Dict:
        parsed = self._parsed = (
            json.loads((await self.blob_async).read().decode("utf-8"))
            if self._parsed is None
            else self._parsed
        )
        if self._parsed is None:
            self._parsed = parsed
        return parsed
